/**** A P P L I C A T I O N   N O T E   ************************************
*
* Title		: DMX512 8ch demultiplexer
* Version		: v2.0
* Last updated	: 17.04.21
* Target		: Transceiver Rev.3.01 [ATmega8515]
* Clock			: 8MHz
*
* written by hendrik hoelscher, www.hoelscher-hi.de with ATMEL Studio 7
***************************************************************************
 This program is free software; you can redistribute it and/or 
 modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation; either version2 of 
 the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 General Public License for more details. 

 If you have no copy of the GNU General Public License, write to the 
 Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

 For other license models, please contact the author.

;***************************************************************************/

#include "lib_indicator.h"
#include "lib_dmx_in.h"

#include <avr/pgmspace.h>
#include <avr/wdt.h>

		

volatile uint8_t   Flags;
volatile uint8_t   PdmCnt;
         uint8_t   PdmCompare[256];
volatile uint8_t   PdmField[8];
		 uint8_t   IndCnt;
		 uint16_t  ChaseCnt;			//low: gradient; high: step
		 uint8_t   StepBuf;
		 
		 
typedef union {
	uint16_t u16;
	struct {
		uint8_t u8l;
		uint8_t u8h;
	};
} uint16or8;


enum {TICK};

#define IND_TIME	(10)
#define BLINK_TIME  (3)

const uint8_t NumToBit[]   ={0b00000001, 0b00000010, 0b00000100, 0b00001000, 0b00010000, 0b00100000, 0b01000000, 0b10000000};

const uint8_t PatRes[] PROGMEM =	{1, 2, 3, 4, 6, 8, 10, 14, 16, 18, 20, 24, 28, 32, 36, 42, 50};

const uint8_t Pattern[] PROGMEM =	{								
									0b11111111,		//pat0
									0b11111111,
									0b11111111,
									0b11111111,
									0b11111111,
									0b11111111,
									0b11111111,
									0b11111111,
									
									0b10101010,		//pat1
									0b01010101,
									0b10101010,
									0b01010101,
									0b10101010,
									0b01010101,
									0b10101010,
									0b01010101,
									
									0b10000000,		//pat2
									0b01000000,
									0b00100000,
									0b00010000,
									0b00001000,
									0b00000100,
									0b00000010,
									0b00000001,
									
									0b10000001,		//pat3
									0b01000010,
									0b00100100,
									0b00011000,
									0b00011000,
									0b00100100,
									0b01000010,
									0b10000001,
									
									0b10000001,		//pat4
									0b01000010,
									0b00100100,
									0b00011000,
									0b00011000,
									0b00111100,
									0b01111110,
									0b11111111,
									
									0b00000001,		//pat5
									0b00000011,
									0b00000111,
									0b00001111,
									0b00011111,
									0b00111111,
									0b01111111,
									0b11111111,
									
									0b00000001,		//pat6
									0b01000000,
									0b00000100,
									0b00010000,
									0b00001000,
									0b00100000,
									0b00000010,
									0b10000000,
									
									0b10000001,		//pat7
									0b00000011,
									0b00000110,
									0b00001100,
									0b00011000,
									0b00110000,
									0b01100000,
									0b11000000
								};




// *************** Initialization ****************
void init_system(void)
{
uint8_t i;
uint8_t tmp;
wdt_reset();
wdt_enable(WDTO_500MS);											//enable watchdog (Timeout= 0.5s)
wdt_reset();

DDRE=  0b00000001;												//LED2 off, BIT9 & OPTION Pullup
PORTE= 0b11111111;

DDRD=  0b10000110;												//DMX, spare, LED1 on
PORTD= 0b01111010;

DDRB=  0b00000000;
PORTB= 0b00000011;

DDRA=  0b11111111;												
if (PIND &(1<<PD4)) PORTA= 0x00;								//low outputs if Spare1 not jumpered
else				PORTA= 0xFF;

TCCR0  = (1<<CS01);												//set timer0 to ck/8
TIMSK  = (1<<TOIE0);											//enable Timer0 overflow interrupt (pulse density modulation, SysTick)

init_ind();
init_DMX_RX();
get_dips();

ChaseCnt= 0;

for (i=0; i<8; i++)
   {
	PdmField[i]= 0;   
   }

// pdm permutation lookup table (see AN014 for more information)
PdmCompare[0x00]= 1;
for (i=1; i!=0xFF; i++)
    {
	tmp= 0;
	if (i &(1<<7)) tmp |= (1<<0);
	if (i &(1<<6)) tmp |= (1<<1);
	if (i &(1<<5)) tmp |= (1<<2);
	if (i &(1<<4)) tmp |= (1<<3);
	if (i &(1<<3)) tmp |= (1<<4);
	if (i &(1<<2)) tmp |= (1<<5);
	if (i &(1<<1)) tmp |= (1<<6);
	if (i &(1<<0)) tmp |= (1<<7);
	PdmCompare[i]= tmp;
    }
PdmCompare[0xFF]= 0xFF;
}


void eval_chase(void)
{
uint8_t res;
uint8_t pat;
uint8_t tmp;
uint16or8 chs;
IndFlags |= (1<<VALID_DMX)|(1<<SIGNAL_COMING);					//don't indicate DMX errors in chaser mode

res = (0xFF -PINC);												//calc resolution from DIPs
res = (res>>4)&0b1111;											//resolution is bit4..7	(bit 3 currently not supported)
res = pgm_read_byte(&PatRes[res]);								//get resolution from lookup table (speed is fixed @100Hz)
chs.u16  = ChaseCnt;
chs.u16 += res;
if (chs.u8h >= 8) chs.u8h=0;									//limit steps to 0..7
ChaseCnt = chs.u16;				

pat = (0xFF -PINC);												//calc base address of pattern from DIPs
pat = (pat &0b111);												//pattern is bit0..2
pat = (pat *8);													//each pattern has 8 steps
pat = (pat +chs.u8h);											//base address +current step
pat= pgm_read_byte(&Pattern[pat]);
if (!(PIND &(1<<PD4))) pat= 0xFF -pat;							//invert pattern if Spare1 set

if (chs.u8h != StepBuf)
   {
	StepBuf= chs.u8h;
	IndFlags |= (1<<DATA_REFRESHED);    
   }

if(PINE &(1<<PE2))
   {
	//fade
	for (tmp=0; tmp<8; tmp++)
		{
		chs.u8l= PdmField[tmp];
		chs.u8h= 0;
		if (pat &NumToBit[tmp])
		   {
		   chs.u16 += res;										//fade up
		   if (chs.u8h != 0) chs.u8l= 0xFF;						//limit on full brightness
		   }
		else
		   {
		   chs.u16 -= res;										//fade down
		   if (chs.u8h != 0) chs.u8l= 0x00;						//limit on black
		   }
		PdmField[tmp]= chs.u8l;
	   }
   }
else
   {
	//switch
	for (tmp=0; tmp<8; tmp++)
		{
		if (pat &NumToBit[tmp]) PdmField[tmp]= 0xFF;
		else                    PdmField[tmp]= 0x00;
		}
	}
}


void eval_dmx(void)
{
uint8_t ch;
uint8_t buf;
if (PIND &(1<<PD4))
   {
	for (ch=0; ch<8; ch++) 
	   {
		buf= DmxRxField[ch];
		if(PdmField[ch] != buf)
		   {
		   PdmField[ch]= buf;
		   IndFlags |= (1<<DATA_REFRESHED);
		   }
	   }
   }
else
   {
	for (ch=0; ch<8; ch++)
	   {
		buf= 0xFF -DmxRxField[ch];
		if(PdmField[ch] != buf)
		   {
			PdmField[ch]= buf;
			IndFlags |= (1<<DATA_REFRESHED);
		   }
	   }
   }
}


int main(void)
{
cli();
init_system();
sei();
for(;;)
	{
	if (Flags &(1<<TICK))						//called with 100Hz
		{
		Flags &= ~(1<<TICK);
		wdt_reset();
		if (PINE &(1<<PE1)) eval_dmx();
		else                eval_chase();
		
		IndCnt--;
		if (IndCnt == 0)
			{
			IndCnt= IND_TIME;
			get_dips();
			indicate();
			}
		}
	};
}



// ****************** PDM compare ISR ********************
ISR (TIMER0_OVF_vect)
{
uint8_t cmp;
uint8_t PdmState= 0;								//output buffer (clear all outputs)
if (++PdmCnt == 0) Flags |= (1<<TICK);
	
cmp= PdmCompare[PdmCnt];							//get pdm compare value
if (PdmField[0] >= cmp) PdmState |= (1<<PA0);
if (PdmField[1] >= cmp) PdmState |= (1<<PA1);
if (PdmField[2] >= cmp) PdmState |= (1<<PA2);
if (PdmField[3] >= cmp) PdmState |= (1<<PA3);
if (PdmField[4] >= cmp) PdmState |= (1<<PA4);
if (PdmField[5] >= cmp) PdmState |= (1<<PA5);
if (PdmField[6] >= cmp) PdmState |= (1<<PA6);
if (PdmField[7] >= cmp) PdmState |= (1<<PA7);
PORTA= PdmState;									//set outputs

TCNT0= 216;											//seed Timer (insert curves here)
}



